@@ -23,15 +23,15 @@
 
 /* Note that BB means BUGBUG (ie something to fix eventually) */
 
-
 #include <linux/module.h>
 #include <linux/fs.h>
+#include <linux/mount.h>
 #include <linux/slab.h>
 #include <linux/init.h>
 #include <linux/version.h>
-#include <linux/locks.h>
 #include <linux/list.h>
 #include <linux/seq_file.h>
+#include <linux/vfs.h>
 #include "cifsfs.h"
 #include "cifspdu.h"
 #define DECLARE_GLOBALS_HERE
@@ -53,27 +53,30 @@ unsigned int extended_security = 0;
 unsigned int ntlmv2_support = 0;
 unsigned int sign_CIFS_PDUs = 0;
 
-extern int cifs_mount(struct super_block *, struct cifs_sb_info *, char *);
+extern int cifs_mount(struct super_block *, struct cifs_sb_info *, char *,
+			char *);
 extern int cifs_umount(struct super_block *, struct cifs_sb_info *);
 void cifs_proc_init(void);
 void cifs_proc_clean(void);
 
-struct super_block *cifs_read_super(struct super_block *sb,
-				    void *data, int silent)
+struct super_block *
+cifs_read_super(struct super_block *sb, void *data, int silent)
 {
 	struct inode *inode;
 	struct cifs_sb_info *cifs_sb;
 	int rc = 0;
 
 	cifs_sb = CIFS_SB(sb);
+	if(cifs_sb == NULL)
+		return 0;
 	cifs_sb->local_nls = load_nls_default();	/* needed for ASCII cp to Unicode converts */
-	rc = cifs_mount(sb, cifs_sb, data);
+
+	rc = cifs_mount(sb, cifs_sb, data, NULL);
 
 	if (rc) {
 		if (!silent)
 			cERROR(1,
-			       ("cifs_mount failed w/return code = %d\n",
-				rc));
+			       ("cifs_mount failed w/return code = %d", rc));
 		goto out_mount_failed;
 	}
 
@@ -94,41 +97,37 @@ struct super_block *cifs_read_super(struct super_block *sb,
 	return sb;
 
 out_no_root:
-	cEVENT(1, ("cifs_read_super: get root inode failed\n"));
+	cERROR(1, ("cifs_read_super: get root inode failed"));
 	if (inode)
 		iput(inode);
 
-/*	rc = cifs_umount(sb);  BB is CIFS unmount routine needed? */
-	if (rc) {
-		cERROR(1, ("cifs_mount failed with no root inode"));
-	}
 out_mount_failed:
-	sb->s_dev = 0;
+    sb->s_dev = 0;
 	return 0;
 }
 
-void cifs_put_super(struct super_block *sb)
+void
+cifs_put_super(struct super_block *sb)
 {
 	int rc = 0;
 	struct cifs_sb_info *cifs_sb;
 
-	cFYI(1, ("In cifs_put_super\n"));
+	cFYI(1, ("In cifs_put_super"));
 	cifs_sb = CIFS_SB(sb);
 	if(cifs_sb == NULL) {
-		cFYI(1,("\nEmpty cifs superblock info passed to unmount"));
+		cFYI(1,("Empty cifs superblock info passed to unmount"));
 		return;
 	}
 	rc = cifs_umount(sb, cifs_sb); 
 	if (rc) {
-		cERROR(1,
-		       ("cifs_umount failed with return code %d\n", rc));
+		cERROR(1, ("cifs_umount failed with return code %d", rc));
 	}
 	unload_nls(cifs_sb->local_nls);
-	cifs_sb->local_nls = NULL;
 	return;
 }
 
-int cifs_statfs(struct super_block *sb, struct statfs *buf)
+int
+cifs_statfs(struct super_block *sb, struct statfs *buf)
 {
 	int xid, rc;
 	struct cifs_sb_info *cifs_sb;
@@ -156,67 +155,203 @@ int cifs_statfs(struct super_block *sb, struct statfs *buf)
 	   __fsid_t f_fsid;
 	   int f_namelen;  */
 	/* BB get from info put in tcon struct at mount time with call to QFSAttrInfo */
-
 	FreeXid(xid);
 	return 0;		/* always return success? what if volume is no longer available? */
 }
 
+kmem_cache_t *cifs_req_cachep;
+kmem_cache_t *cifs_mid_cachep;
+
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0))
+static struct inode *
+cifs_alloc_inode(struct super_block *sb)
+{
+	struct cifsInodeInfo *cifs_inode;
+	cifs_inode =
+	    (struct cifsInodeInfo *) kmem_cache_alloc(cifs_inode_cachep,
+						      SLAB_KERNEL);
+	if (!cifs_inode)
+		return NULL;
+	cifs_inode->cifsAttrs = 0x20;	/* default */
+	atomic_set(&cifs_inode->inUse, 0);
+	cifs_inode->time = 0;
+    if(oplockEnabled)
+	    cifs_inode->clientCanCacheAll = 1;
+	INIT_LIST_HEAD(&cifs_inode->openFileList);
+	return &cifs_inode->vfs_inode;
+}
+
+static void
+cifs_destroy_inode(struct inode *inode)
+{
+	kmem_cache_free(cifs_inode_cachep, CIFS_I(inode));
+}
+#endif
+
+/*
+ * cifs_show_options() is for displaying mount options in /proc/mounts.
+ * It tries to avoid showing settings that were not changed from their
+ * defaults.
+ */
+static int
+cifs_show_options(struct seq_file *s, struct vfsmount *m)
+{
+	struct cifs_sb_info *cifs_sb;
+
+	cifs_sb = CIFS_SB(m->mnt_sb);
+
+	if (cifs_sb)
+		if (cifs_sb->tcon) {
+			seq_printf(s, ", TARGET: %s ", cifs_sb->tcon->treeName);
+			seq_printf(s, "FS TYPE: %s ",
+				   cifs_sb->tcon->nativeFileSystem);
+			if (cifs_sb->tcon->ses->userName)
+				seq_printf(s, " USER: %s ",
+					   cifs_sb->tcon->ses->userName);
+		}
+	return 0;
+}
+
 struct super_operations cifs_super_ops = {
-	read_inode:cifs_read_inode,
-	put_super:cifs_put_super,
-	statfs:cifs_statfs,
-	/*  umount_begin: cifs_umount_begin,
-	   dirty_inode:  cifs_dirty_inode,
-   remount_fs:   cifs_remount_fs, *//* TBD */
+	.read_inode = cifs_read_inode,
+	.put_super = cifs_put_super,
+	.statfs = cifs_statfs,
+/*	.alloc_inode = cifs_alloc_inode,
+	.destroy_inode = cifs_destroy_inode, */
+/*	.drop_inode	    = generic_delete_inode, 
+	.delete_inode	= cifs_delete_inode,  *//* Do not need the above two functions     
+   unless later we add lazy close of inodes or unless the kernel forgets to call
+   us with the same number of releases (closes) as opens */
+	.show_options = cifs_show_options,
+/*    .umount_begin   = cifs_umount_begin, *//* consider adding in the future */
 };
 
-static DECLARE_FSTYPE(cifs_fs_type, "cifs", cifs_read_super,
-		      0 /* BUGBUG FS_ flags go here */ );
-struct inode_operations cifs_dir_inode_ops = {	/* divide the following into file vs. dir inode operations */
-	create:cifs_create,
-	lookup:cifs_lookup,
-	unlink:cifs_unlink,
-	link:cifs_hardlink,
-	mkdir:cifs_mkdir,
-	rmdir:cifs_rmdir,
-	rename:cifs_rename,
-	revalidate:cifs_revalidate,
-	setattr:cifs_setattr,
-	symlink:cifs_symlink,
+#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0))
+static struct super_block *
+cifs_get_sb(struct file_system_type *fs_type,
+	    int flags, char *dev_name, void *data)
+{
+	int rc;
+	struct super_block *sb = sget(fs_type, NULL, set_anon_super, NULL);
+
+	cFYI(1, ("Devname: %s flags: %d ", dev_name, flags));
+
+	if (IS_ERR(sb))
+		return sb;
+
+	sb->s_flags = flags;
+
+	rc = cifs_read_super(sb, data, dev_name, flags & MS_VERBOSE ? 1 : 0);
+	if (rc) {
+		up_write(&sb->s_umount);
+		deactivate_super(sb);
+		return ERR_PTR(rc);
+	}
+	sb->s_flags |= MS_ACTIVE;
+	return sb;
+}
+
+static struct file_system_type cifs_fs_type = {
+	.owner = THIS_MODULE,
+	.name = "cifs",
+	.get_sb = cifs_get_sb,
+	.kill_sb = kill_anon_super,
+	/*  .fs_flags */
+};
+#endif
+
+static DECLARE_FSTYPE(cifs_fs_type, "cifs", cifs_read_super,0);
+
+
+struct inode_operations cifs_dir_inode_ops = {
+	.create = cifs_create,
+	.lookup = cifs_lookup,
+	.unlink = cifs_unlink,
+	.link = cifs_hardlink,
+	.mkdir = cifs_mkdir,
+	.rmdir = cifs_rmdir,
+	.rename = cifs_rename,
+	.revalidate = cifs_revalidate,  
+	.setattr = cifs_setattr,
+	.symlink = cifs_symlink,
 };
 
 struct inode_operations cifs_file_inode_ops = {
-	revalidate:cifs_revalidate,
-	setattr:cifs_setattr,
-	rename:cifs_rename,
+	.revalidate = cifs_revalidate,
+	.setattr = cifs_setattr,
+/*	.getattr = cifs_getattr,*/
+	.rename = cifs_rename,
+#ifdef CIFS_XATTR
+	.setxattr = cifs_setxattr,
+	.getxattr = cifs_getxattr,
+	.listxattr = cifs_listxattr,
+	.removexattr = cifs_removexattr,
+	.permission = cifs_permission,
+#endif 
 };
 
 struct inode_operations cifs_symlink_inode_ops = {
-	readlink:cifs_readlink,
-	follow_link:cifs_follow_link,
+	.readlink = cifs_readlink,
+	.follow_link = cifs_follow_link,
 	/* BB add the following two eventually */
-	/* revalidate:     cifs_revalidate,
-	   setattr:        cifs_notify_change, */
+	/* revalidate: cifs_revalidate,
+	   setattr:    cifs_notify_change, *//* BB do we need notify change */
+#ifdef CIFS_XATTR
+	.setxattr = cifs_setxattr,
+	.getxattr = cifs_getxattr,
+	.listxattr = cifs_listxattr,
+	.removexattr = cifs_removexattr,
+#endif 
 };
 
 struct file_operations cifs_file_ops = {
-	read:generic_file_read,
-	write:generic_file_write,
-	open:cifs_open,
-	release:cifs_close,
-	lock:cifs_lock,
-	fsync:cifs_fsync,
-	mmap:cifs_file_mmap,
+	.read = generic_file_read,
+	.write = generic_file_write, 
+	.open = cifs_open,
+	.release = cifs_close,
+	.lock = cifs_lock,
+	.fsync = cifs_fsync,
+    .flush = cifs_flush,
+	.mmap  = cifs_file_mmap,
+/*	.sendfile = generic_file_sendfile,*/
 };
 
 struct file_operations cifs_dir_ops = {
-/* open:        cifs_dir_open, *//* turns out to be unneeded - we FindFirst, FindNext and close search in cifs_readdir */
-	readdir:cifs_readdir,
-	release:cifs_closedir,
+	.readdir = cifs_readdir,
+	.release = cifs_closedir,
 };
+/*
+static void
+cifs_init_once(void *inode, kmem_cache_t * cachep, unsigned long flags)
+{
+	struct cifsInodeInfo *cifsi = (struct cifsInodeInfo *) inode;
 
-kmem_cache_t *cifs_req_cachep;
-kmem_cache_t *cifs_mid_cachep;
+	if ((flags & (SLAB_CTOR_VERIFY | SLAB_CTOR_CONSTRUCTOR)) ==
+	    SLAB_CTOR_CONSTRUCTOR) {
+		inode_init_once(&cifsi->vfs_inode);
+		INIT_LIST_HEAD(&cifsi->lockList);
+	}
+}
+
+int
+cifs_init_inodecache(void)
+{
+	cifs_inode_cachep = kmem_cache_create("cifs_inode_cache",
+					      sizeof (struct cifsInodeInfo),
+					      0, SLAB_HWCACHE_ALIGN,
+					      cifs_init_once, NULL);
+	if (cifs_inode_cachep == NULL)
+		return -ENOMEM;
+
+	return 0;
+}
+
+void
+cifs_destroy_inodecache(void)
+{
+	if (kmem_cache_destroy(cifs_inode_cachep))
+		printk(KERN_WARNING "cifs_inode_cache: error freeing\n");
+} */
 
 int
 cifs_init_request_bufs(void)
@@ -235,7 +370,7 @@ void
 cifs_destroy_request_bufs(void)
 {
 	if (kmem_cache_destroy(cifs_req_cachep))
-		printk(KERN_INFO
+		printk(KERN_WARNING
 		       "cifs_destroy_request_cache: error not all structures were freed\n");
 }
 
@@ -255,11 +390,12 @@ void
 cifs_destroy_mids(void)
 {
 	if (kmem_cache_destroy(cifs_mid_cachep))
-		printk(KERN_INFO
+		printk(KERN_WARNING
 		       "cifs_destroy_mids: error not all structures were freed\n");
 }
 
-static int __init init_cifs(void)
+static int __init
+init_cifs(void)
 {
 	int rc = 0;
 #if CONFIG_PROC_FS
@@ -282,36 +418,43 @@ static int __init init_cifs(void)
 	GlobalSMBSeslock = RW_LOCK_UNLOCKED;
 	GlobalMid_Lock = RW_LOCK_UNLOCKED;
 
-	rc = cifs_init_mids();
+/*	rc = cifs_init_inodecache();*/
 	if (!rc) {
-		rc = cifs_init_request_bufs();
+		rc = cifs_init_mids();
 		if (!rc) {
-			rc = register_filesystem(&cifs_fs_type);
-			if (!rc)
-				return rc;	/* Success */
-			else
-				cifs_destroy_request_bufs();
+			rc = cifs_init_request_bufs();
+			if (!rc) {
+				rc = register_filesystem(&cifs_fs_type);
+				if (!rc)
+					return rc;	/* Success */
+				else
+					cifs_destroy_request_bufs();
+			}
+			cifs_destroy_mids();
 		}
-		cifs_destroy_mids();
+/*		cifs_destroy_inodecache(); */
 	}
-	return rc;
-}
-
-static void __exit exit_cifs(void)
-{
-	cFYI(0, ("\nIn unregister ie exit_cifs"));
 #if CONFIG_PROC_FS
 	cifs_proc_clean();
 #endif
-	unregister_filesystem(&cifs_fs_type);
+	return rc;
+}
+
+static void __exit
+exit_cifs(void)
+{
+	cFYI(0, ("In unregister ie exit_cifs"));
+#if CONFIG_PROC_FS
+	cifs_proc_clean();
+#endif
+    	unregister_filesystem(&cifs_fs_type);
+/*	cifs_destroy_inodecache();*/
 	cifs_destroy_mids();
 	cifs_destroy_request_bufs();
 }
 
 MODULE_AUTHOR("Steve French <sfrench@us.ibm.com>");
-#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,4,7))
 MODULE_LICENSE("GPL");		/* combination of LGPL + GPL source behaves as GPL */
-#endif
 MODULE_DESCRIPTION
     ("VFS to access servers complying with the SNIA CIFS Specification e.g. Samba and Windows");
 module_init(init_cifs)
