@@ -34,7 +34,6 @@ cifs_get_inode_info_unix(struct inode **pinode,
 			 struct super_block *sb)
 {
 	int xid;
-	int newinode = FALSE;
 	int rc = 0;
 	FILE_UNIX_BASIC_INFO findData;
 	struct cifsTconInfo *pTcon;
@@ -45,7 +44,7 @@ cifs_get_inode_info_unix(struct inode **pinode,
 	xid = GetXid();
 
 	pTcon = cifs_sb->tcon;
-	cFYI(1, ("\nGetting info on %s ", search_path));
+	cFYI(1, (" Getting info on %s ", search_path));
 	/* we could have done a find first instead but this returns more info */
 	rc = CIFSSMBUnixQPathInfo(xid, pTcon, search_path, &findData,
 				  cifs_sb->local_nls);
@@ -86,24 +85,17 @@ cifs_get_inode_info_unix(struct inode **pinode,
 		/* get new inode */
 		if (*pinode == NULL) {
 			*pinode = new_inode(sb);
-			cFYI(1, ("\nAlloc new inode %p ", *pinode));
-			if(*pinode)  {
-				newinode = TRUE;
-			}
+			cFYI(1, (" Alloc new inode %p ", *pinode));
 		}
 		inode = *pinode;
-/*      new_inode = iget(parent_dir_inode->i_sb, findData.IndexNumber); */
+/*        new_inode = iget(parent_dir_inode->i_sb, findData.IndexNumber); */
         /* index number not reliable in response data */
 
 		cifsInfo = CIFS_I(inode);
 
-		cFYI(1, ("\nOld time %ld ", cifsInfo->time));
+		cFYI(1, (" Old time %ld ", cifsInfo->time));
 		cifsInfo->time = jiffies;
 		cFYI(1, (" New time %ld ", cifsInfo->time));
-		if(newinode == TRUE) {
-			cifsInfo->clientCanCache = 0;
-			INIT_LIST_HEAD(&cifsInfo->openFileList);
-		}
 		atomic_inc(&cifsInfo->inUse);	/* inc on every refresh of inode */
 
 		inode->i_atime =
@@ -137,12 +129,12 @@ cifs_get_inode_info_unix(struct inode **pinode,
 		findData.EndOfFile = le64_to_cpu(findData.EndOfFile);
 		inode->i_size = findData.EndOfFile;
 		inode->i_blksize =
-		    (pTcon->ses->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;
+		    (pTcon->ses->server->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;
 		inode->i_blocks = do_div(findData.NumOfBytes, inode->i_blksize);
 		if (findData.NumOfBytes < findData.EndOfFile)
-			cFYI(1, ("\nServer inconsistency Error: it says allocation size less than end of file "));
+			cFYI(1, ("Server inconsistency Error: it says allocation size less than end of file "));
 		cFYI(1,
-		     ("\nCIFS FFIRST: Size %ld and blocks %ld ",
+		     ("Size %ld and blocks %ld ",
 		      (unsigned long) inode->i_size, inode->i_blocks));
 		if (S_ISREG(inode->i_mode)) {
 			cFYI(1, (" File inode "));
@@ -158,7 +150,7 @@ cifs_get_inode_info_unix(struct inode **pinode,
 			inode->i_op = &cifs_symlink_inode_ops;
 /* tmp_inode->i_fop = *//* do not need to set to anything */
 		} else {
-			cFYI(1, ("\nInit special inode "));
+			cFYI(1, (" Init special inode "));
 			init_special_inode(inode, inode->i_mode,
 					   kdev_t_to_nr(inode->i_rdev));
 		}
@@ -173,7 +165,6 @@ cifs_get_inode_info(struct inode **pinode,
 {
 	int xid;
 	int rc = 0;
-	int newinode = FALSE;
 	FILE_ALL_INFO findData;
 	struct cifsTconInfo *pTcon;
 	struct inode *inode;
@@ -183,7 +174,7 @@ cifs_get_inode_info(struct inode **pinode,
 	xid = GetXid();
 
 	pTcon = cifs_sb->tcon;
-	cFYI(1, ("\nGetting info on %s ", search_path));
+	cFYI(1, (" Getting info on %s ", search_path));
 	/* we could have done a find first instead but this returns more info */
 	rc = CIFSSMBQPathInfo(xid, pTcon, search_path, &findData,
 			      cifs_sb->local_nls);
@@ -223,27 +214,20 @@ cifs_get_inode_info(struct inode **pinode,
 		/* get new inode */
 		if (*pinode == NULL) {
 			*pinode = new_inode(sb);
-			cFYI(1, ("\nAlloc new inode %p ", *pinode));
-			if(*pinode) {
-				newinode = TRUE;
-			}
+			cFYI(1, (" Alloc new inode %p ", *pinode));
 		}
 
 		inode = *pinode;
 		cifsInfo = CIFS_I(inode);
 		findData.Attributes = le32_to_cpu(findData.Attributes);
 		cifsInfo->cifsAttrs = findData.Attributes;
-		cFYI(1, ("\nOld time %ld ", cifsInfo->time));
+		cFYI(1, (" Old time %ld ", cifsInfo->time));
 		cifsInfo->time = jiffies;
 		cFYI(1, (" New time %ld ", cifsInfo->time));
-		if(newinode == TRUE) {
-			cifsInfo->clientCanCache = 0;
-			INIT_LIST_HEAD(&cifsInfo->openFileList);
-		}
 		atomic_inc(&cifsInfo->inUse);	/* inc on every refresh of inode */
 
 		inode->i_blksize =
-		    (pTcon->ses->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;
+		    (pTcon->ses->server->maxBuf - MAX_CIFS_HDR_SIZE) & 0xFFFFFE00;
 		/* Linux can not store file creation time unfortunately so we ignore it */
 		inode->i_atime =
 		    cifs_NTtimeToUnix(le64_to_cpu(findData.LastAccessTime));
@@ -255,7 +239,7 @@ cifs_get_inode_info(struct inode **pinode,
 		/* should we treat the dos attribute of read-only as read-only mode bit e.g. 555 */
 		inode->i_mode = S_IALLUGO & ~(S_ISUID | S_IXGRP);	/* 2767 perms indicate mandatory locking - will override for dirs later */
 		cFYI(0,
-		     ("\nAttributes came in as 0x%x\n", findData.Attributes));
+		     (" Attributes came in as 0x%x ", findData.Attributes));
 		if (findData.Attributes & ATTR_REPARSE) {	
    /* Can IFLNK be set as it basically is on windows with IFREG or IFDIR? */
 			inode->i_mode |= S_IFLNK;
@@ -272,7 +256,7 @@ cifs_get_inode_info(struct inode **pinode,
 		inode->i_blocks =
 		    do_div(findData.AllocationSize, inode->i_blksize);
 		cFYI(1,
-		     ("\n Size %ld and blocks %ld ",
+		     (" Size %ld and blocks %ld ",
 		      (unsigned long) inode->i_size, inode->i_blocks));
 		inode->i_nlink = le32_to_cpu(findData.NumberOfLinks);
 
@@ -323,21 +307,31 @@ cifs_unlink(struct inode *inode, struct dentry *direntry)
 	char *full_path = NULL;
 	struct cifsInodeInfo *cifsInode;
 
-	cFYI(1, ("\n cifs_unlink, inode = 0x%p with ", inode));
+	cFYI(1, (" cifs_unlink, inode = 0x%p with ", inode));
 
 	xid = GetXid();
 
 	cifs_sb = CIFS_SB(inode->i_sb);
 	pTcon = cifs_sb->tcon;
 
-	/* BB Should we close the file if it is already open from our client? */
-
 	full_path = build_path_from_dentry(direntry);
 
 	rc = CIFSSMBDelFile(xid, pTcon, full_path, cifs_sb->local_nls);
 
 	if (!rc) {
 		direntry->d_inode->i_nlink--;
+	} else if (rc == -ETXTBSY) {
+		int oplock = FALSE;
+		__u16 netfid;
+
+		rc = CIFSSMBOpen(xid, pTcon, full_path, FILE_OPEN, DELETE, 
+				CREATE_NOT_DIR | CREATE_DELETE_ON_CLOSE,
+				&netfid, &oplock, cifs_sb->local_nls);
+		if(rc==0) {
+			CIFSSMBClose(xid, pTcon, netfid);
+			/* BB In the future chain close with the NTCreateX to narrow window */
+			direntry->d_inode->i_nlink--;
+		}
 	}
 	cifsInode = CIFS_I(direntry->d_inode);
 	cifsInode->time = 0;	/* will force revalidate to get info when needed */
@@ -362,7 +356,7 @@ cifs_mkdir(struct inode *inode, struct dentry *direntry, int mode)
 	char *full_path = NULL;
 	struct inode *newinode = NULL;
 
-	cFYI(1, ("In cifs_mkdir, mode = 0x%x inode = 0x%p\n", mode, inode));
+	cFYI(1, ("In cifs_mkdir, mode = 0x%x inode = 0x%p ", mode, inode));
 
 	xid = GetXid();
 
@@ -373,7 +367,7 @@ cifs_mkdir(struct inode *inode, struct dentry *direntry, int mode)
 	/* BB add setting the equivalent of mode via CreateX w/ACLs */
 	rc = CIFSSMBMkDir(xid, pTcon, full_path, cifs_sb->local_nls);
 	if (rc) {
-		cFYI(1, ("\ncifs_mkdir returned 0x%x ", rc));
+		cFYI(1, ("cifs_mkdir returned 0x%x ", rc));
 	} else {
 		inode->i_nlink++;
 		if (pTcon->ses->capabilities & CAP_UNIX)
@@ -413,7 +407,7 @@ cifs_rmdir(struct inode *inode, struct dentry *direntry)
 	char *full_path = NULL;
 	struct cifsInodeInfo *cifsInode;
 
-	cFYI(1, ("\nn cifs_rmdir, inode = 0x%p with ", inode));
+	cFYI(1, (" cifs_rmdir, inode = 0x%p with ", inode));
 
 	xid = GetXid();
 
@@ -462,14 +456,19 @@ cifs_rename(struct inode *source_inode, struct dentry *source_direntry,
 	if (pTcon != cifs_sb_target->tcon) {    
 		return -EXDEV;	/* BB actually could be allowed if same server, but
                      different share. Might eventually add support for this */
-        FreeXid(xid);
-    }
+        	FreeXid(xid);
+	}
 
 	fromName = build_path_from_dentry(source_direntry);
 	toName = build_path_from_dentry(target_direntry);
 
 	rc = CIFSSMBRename(xid, pTcon, fromName, toName,
 			   cifs_sb_source->local_nls);
+	if(rc == -EEXIST) {
+		cifs_unlink(target_inode, target_direntry);
+		rc = CIFSSMBRename(xid, pTcon, fromName, toName,
+				   cifs_sb_source->local_nls);
+	}
 	if (fromName)
 		kfree(fromName);
 	if (toName)
@@ -494,20 +493,25 @@ cifs_revalidate(struct dentry *direntry)
 
 	full_path = build_path_from_dentry(direntry);
 	cFYI(1,
-	     (" full path: %s for inode 0x%p with count %d dentry: 0x%p d_time %ld at time %ld \n",
+	     ("Revalidate full path: %s for inode 0x%p with count %d dentry: 0x%p d_time %ld at time %ld ",
 	      full_path, direntry->d_inode,
 	      direntry->d_inode->i_count.counter, direntry,
 	      direntry->d_time, jiffies));
 
 	cifsInode = CIFS_I(direntry->d_inode);
-/* BB add check - do not need to revalidate oplocked files */
-	if ((time_before(jiffies, cifsInode->time + HZ))
-	    && (direntry->d_inode->i_nlink == 1)) {
-		cFYI(1, (" Do not need to revalidate "));
-		if (full_path)
-			kfree(full_path);
-		FreeXid(xid);
-		return rc;
+	/* BB add check - do not need to revalidate oplocked files */
+
+	if (time_before(jiffies, cifsInode->time + HZ)) {
+	    if((S_ISREG(direntry->d_inode->i_mode) == 0) || 
+              (direntry->d_inode->i_nlink == 1)) {
+		    if (full_path)
+			    kfree(full_path);
+		    FreeXid(xid);
+		    return rc;
+        } else {
+            cFYI(1,("Have to revalidate file due to hardlinks"));
+        }
+            
 	}
 
 	if (cifs_sb->tcon->ses->capabilities & CAP_UNIX)
@@ -526,6 +530,14 @@ cifs_revalidate(struct dentry *direntry)
 	return rc;
 }
 
+/* int cifs_getattr(struct vfsmount *mnt, struct dentry *dentry, struct kstat *stat)
+{
+	int err = cifs_revalidate(dentry);
+	if (!err)
+		generic_fillattr(dentry->d_inode, stat);
+	return err;
+} */
+
 void
 cifs_truncate_file(struct inode *inode)
 {				/* BB remove - may not need this function after all BB */
@@ -555,7 +567,7 @@ cifs_truncate_file(struct inode *inode)
 		full_path = build_path_from_dentry(dirent);
 		rc = CIFSSMBSetEOF(xid, pTcon, full_path, inode->i_size,FALSE,
 				   cifs_sb->local_nls);
-		cFYI(1,("\nSetEOF (truncate) rc = %d",rc));
+		cFYI(1,(" SetEOF (truncate) rc = %d",rc));
 		if(rc == -ETXTBSY) {        
 			cifsInode = CIFS_I(inode);
 			if(!list_empty(&(cifsInode->openFileList))) {            
@@ -565,7 +577,7 @@ cifs_truncate_file(struct inode *inode)
 				 rc = CIFSSMBSetFileSize(xid, pTcon, inode->i_size,
 					open_file->netfid,open_file->pid,FALSE);
 			} else {
-				  cFYI(1,("\nNo open files to get file handle from"));
+				  cFYI(1,(" No open files to get file handle from"));
 			}
 		}
 		if (!rc)
@@ -597,7 +609,7 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 	xid = GetXid();
 
 	cFYI(1,
-	     ("\nIn cifs_setattr, name = %s attrs->iavalid 0x%x\n",
+	     (" In cifs_setattr, name = %s attrs->iavalid 0x%x ",
 	      direntry->d_name.name, attrs->ia_valid));
 	cifs_sb = CIFS_SB(direntry->d_inode->i_sb);
 	pTcon = cifs_sb->tcon;
@@ -607,12 +619,12 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 
 	/* BB check if we need to refresh inode from server now ? BB */
 
-	cFYI(1, ("\nChanging attributes 0x%x", attrs->ia_valid));
+	cFYI(1, (" Changing attributes 0x%x", attrs->ia_valid));
 
 	if (attrs->ia_valid & ATTR_SIZE) {
 		rc = CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size,FALSE,
 				   cifs_sb->local_nls);
-		cFYI(1,("\nSetEOF (setattrs) rc = %d",rc));
+		cFYI(1,(" SetEOF (setattrs) rc = %d",rc));
 
 		if(rc == -ETXTBSY) {
 			if(!list_empty(&(cifsInode->openFileList))) {            
@@ -622,29 +634,31 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 				rc = CIFSSMBSetFileSize(xid, pTcon, attrs->ia_size,
 					   open_file->netfid,open_file->pid,FALSE);           
 			} else {
-				cFYI(1,("\nNo open files to get file handle from"));
+				cFYI(1,(" No open files to get file handle from"));
 			}
 		}
-        /*  Set Allocation Size of file - might not even need to call the
-            following but might as well and it does not hurt if it fails */
-		CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size, TRUE, cifs_sb->local_nls);
-		if (rc == 0)
-			vmtruncate(direntry->d_inode, attrs->ia_size);
-		/* BB add special case to handle sharing violation (due to Samba bug)
-		   by calling SetFileInfo to set the sizes */
+        /*  For Allocation Size - do not need to call the following
+            it did not hurt if it fails but why bother */
+	/*	CIFSSMBSetEOF(xid, pTcon, full_path, attrs->ia_size, TRUE, cifs_sb->local_nls);*/
+		if (rc == 0) {
+			rc = vmtruncate(direntry->d_inode, attrs->ia_size);
+            /* BB do I need the call below in 2.4 as in 2.5 kernel? */
+	 /*   	nobh_truncate_page(direntry->d_inode->i_mapping, direntry->d_inode->i_size);*/
+/*          cFYI(1,("truncate_page to 0x%lx \n",direntry->d_inode->i_size)); */
+		}
 	}
 	if (attrs->ia_valid & ATTR_UID) {
-		cFYI(1, ("\nCIFS - UID changed to %d", attrs->ia_uid));
+		cFYI(1, (" CIFS - UID changed to %d", attrs->ia_uid));
 		uid = attrs->ia_uid;
 		/*        entry->uid = cpu_to_le16(attr->ia_uid); */
 	}
 	if (attrs->ia_valid & ATTR_GID) {
-		cFYI(1, ("\nCIFS - GID changed to %d", attrs->ia_gid));
+		cFYI(1, (" CIFS - GID changed to %d", attrs->ia_gid));
 		gid = attrs->ia_gid;
 		/*      entry->gid = cpu_to_le16(attr->ia_gid); */
 	}
 	if (attrs->ia_valid & ATTR_MODE) {
-		cFYI(1, ("\nCIFS - Mode changed to 0x%x", attrs->ia_mode));
+		cFYI(1, (" CIFS - Mode changed to 0x%x", attrs->ia_mode));
 		mode = attrs->ia_mode;
 		/*      entry->mode = cpu_to_le16(attr->ia_mode); */
 	}
@@ -673,7 +687,7 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 
 	if (attrs->ia_valid & ATTR_CTIME) {
 		set_time = TRUE;
-		cFYI(1, ("\nCIFS - CTIME changed ")); /* BB probably do not need */
+		cFYI(1, (" CIFS - CTIME changed ")); /* BB probably do not need */
 		time_buf.ChangeTime =
 		    cpu_to_le64(cifs_UnixTimeToNT(attrs->ia_ctime));
 	} else
@@ -688,7 +702,7 @@ cifs_setattr(struct dentry *direntry, struct iattr *attrs)
 		rc = CIFSSMBSetTimes(xid, pTcon, full_path, &time_buf,
 				     cifs_sb->local_nls);
 	}
-	cifsInode->time = 0; /* force revalidate to get attributes when needed */
+/*	cifsInode->time = 0; */ /* force revalidate to get attributes when needed */
 
 	if (full_path)
 		kfree(full_path);
@@ -700,7 +714,7 @@ void
 cifs_delete_inode(struct inode *inode)
 {
 	/* Note: called without the big kernel filelock - remember spinlocks! */
-	cFYI(1, ("In cifs_delete_inode, inode = 0x%p\n", inode));
+	cFYI(1, ("In cifs_delete_inode, inode = 0x%p ", inode));
 	/* may have to add back in when safe distributed caching of
              directories via e.g. FindNotify added */
 }
